/***********************************************************************
BubeView: little software for 3D visualization completely implemented
	  in-software using only the X-window primitives.

Copyright (C) 2005 Andrea Sottoriva 

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
***********************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#define WINSIZE 600
#define WIDTH   600
#define HEIGHT  600

#define TETA_STEP   0.04
#define FI_STEP     0.04
#define ZOOM_STEP   1.00
#define VU_STEP     0.10
#define R_STEP      0.02

#define INSIDE   0
#define CLIPPED  1
#define ZEROFACE 1

typedef enum _view { WIREFRAME, DEPTH_CUING } view_t;

/* punto 2D */
typedef struct _point {
  int x;
  int y;
} point_t;

/* vertice */
typedef struct _vertex {
  float x;
  float y;
  float z;
  char  t;
} vertex_t;

/* faccia triangolare (come previsto dal formato .m) */
typedef struct _face {
  int v1;
  int v2;
  int v3;
  int flag;
} face_t;

/* lato */
typedef struct _edge {
  int p1;
  int p2;
} edge_t;

/* Dati per visualizzare un oggetto definito da vertici, facce e lati */
vertex_t *vertex=NULL;		/* lista vertici */
vertex_t *evert=NULL;
face_t   *face=NULL;		/* lista facce */
edge_t   *edge=NULL;		/* lista lati */
point_t	 *fb=NULL;		/* lista dei punti proiettati sulla window */
vertex_t viewup;		/* view-up vector */

int   nvert, nface, nedge;      /* numero vertici, facce e lati */
float csx, csy, csz;		/* centro oggetto */
float D=1000;			/* distanza dell'osservatore dal centro O */
float teta=2.0;			/* teta dell'osservatore */
float fi=M_PI/2;		/* fi dell'osservatore */
float di=800;			/* distanza dalla window */
float effe=4000;		/* distanza dal far-plane del frustum */
float semw;			/* semilato della window */
float vu_fi=0;			/* fi del view-up vector */
float dstep;			/* step di avanzamento di D */
float sstep;			/* step di spostamento laterale */
float *zbuf;			/* array dei valori di z per il depth-cuing */

int   xvmin, xvmax, yvmin, yvmax; /* coordinate viewport */
float salpha=0.05;		  /* zoomin e zoomout */
float ssx, ssy, ssz, ssmod;	  /* modifica centro oggetto */
float alpha;       /* apertura angolare piramide di vista */
float xwmin, xwmax, ywmin, ywmax; /* coordinate window */
float Sx, Sy;			  /* fattori di scala trasf. window-viewport */

float   tmpx, tmpy;		  /* variabili temporanee */
int     i, j, first=1;

unsigned long scale[256];	/* scala di colori per il depth-cuing */

/* Dati per la costruzione delle finestre */
Display  *mydisp;
Window   mywin;
GC       mygc;
XEvent   myevent;
KeySym   mykey;
XSizeHints   myhint;
Colormap     cmap;
int      myscreen;
unsigned long   myfore, myback;
int      done=0;

char    *filename;
view_t  viewtype;	/* tipo di vista (WIREFRAME | DEPTH_CUING) */

void load_object(void);
void world2view(vertex_t *vin, float matrix[4][4], vertex_t *vout);
void world2view_normal(vertex_t *vin, vertex_t *vout);
void fattori_scala(void);
void draw(void);
void draw_wireframe(void);
void draw_depthcuing(void);
void draw_hiddenline(void);
void clip(float matrix[4][4]);

/* funzioni ausiliarie */
int   round(float g);
int   det(float m[3][3]);
float zminval(face_t f);
float sqr(float x);
float mod(vertex_t *v);
void  dotprod(vertex_t *vin, float matrix[4][4], vertex_t *vout);
void  crossprod(vertex_t *v1, vertex_t *v2, vertex_t *vout);
void  vecdiv(vertex_t *vin, float scalar, vertex_t *vout);
void  spherical2cartesian(float D, float teta, float fi, vertex_t *vout);
void  cartesian2spherical(vertex_t *vin, float *D, float *teta, float *fi);
void  make_perspmatrix(float D, float teta, float fi, float matrix[4][4]);
void  createcolors(void);

int main(int argc, char **argv)
{
  int oldx, oldy, button;
  float pseudoteta, pseudofi;
  char keyseq[10];

  if(argc < 2) {
    fprintf(stderr, "Utilizzo : bubeview <file.m>\n");
    exit(-1);
  }

  printf("\nCOMANDI : \n");
  printf("spostamento laterale mouse                 +-teta\n");
  printf("spostamento verticale mouse                +-fi\n");
  printf("rotella mouse                              +-fi del view-up vector\n");
  printf("arrow up/down/left/right                   spostamento della window\n");
  printf("tasto destro mouse                         micronanimazione sbubba\n\n");
  printf("a         +D\n");
  printf("s         -D\n");
  printf("+         zoom in\n");
  printf("-         zoom out\n\n");
  printf("VISUALIZZAZIONI : \n");
  printf("1         wireframe\n");
  printf("2         depth-cuing\n");

  /* file di mesh (formato .m) */
  filename=argv[1];

  /* costruisco il view-up vector */
  spherical2cartesian(1, M_PI/2, 0, &viewup);
    
  if((mydisp = XOpenDisplay("")) == NULL) {
    fprintf(stderr, "Cannot open display\n");
    exit(-1);
  }

  cmap=DefaultColormap(mydisp, 0);
  myscreen = DefaultScreen(mydisp);
  myfore = WhitePixel(mydisp,myscreen);
  myback = BlackPixel(mydisp,myscreen);
  xvmin = myhint.x = 10;
  yvmin = myhint.y = 10;
  xvmax = myhint.width = myhint.min_width = myhint.max_width = WINSIZE;
  yvmax = myhint.height = myhint.min_height = myhint.max_height = WINSIZE;
  myhint.flags = USPosition | PSize | PMinSize | PMaxSize;
  mywin = XCreateSimpleWindow(mydisp,DefaultRootWindow(mydisp), myhint.x, myhint.y,
			      myhint.width,myhint.height,5,myfore,myback);
  XSetStandardProperties(mydisp,mywin,"Bubeview","Bubeview",None,None,None,&myhint);
  mygc = XCreateGC(mydisp,mywin,0,0);
  XSetBackground(mydisp,mygc,myback);
  XSetForeground(mydisp,mygc,myfore);
  XSelectInput(mydisp,mywin,ButtonPressMask | KeyPressMask | ExposureMask | Button1MotionMask | Button3MotionMask);

  XMapRaised(mydisp,mywin);

  /* alloco la scala di grigi per il depth-cuing */
  createcolors();

  while (done == 0) {
    XNextEvent(mydisp,&myevent);
    switch(myevent.type) {
    case Expose:
      if (myevent.xexpose.count == 0) {
	draw();
      }
      break;
    case MappingNotify:
      XRefreshKeyboardMapping((XMappingEvent*)&myevent);
      break;
    case MotionNotify:
      if(button == 1) {       /* movimento del tasto sinistro del mouse, modifica del punto di vista */
	tmpx = myevent.xbutton.x;
	tmpy = myevent.xbutton.y;
	if(tmpx > oldx) teta-=TETA_STEP;
	if(tmpx < oldx) teta+=TETA_STEP;
	if(tmpy > oldy) {
	  if(fi < M_PI && fi+FI_STEP > M_PI) {
	    vu_fi+=M_PI;
	    spherical2cartesian(1, M_PI/2, vu_fi, &viewup);
	  }
	  fi+=FI_STEP;
	  if(fi >= 2*M_PI) {
	    fi=FI_STEP;
	    vu_fi+=M_PI;
	    spherical2cartesian(1, M_PI/2, vu_fi, &viewup);
	  }
	}
	if(tmpy < oldy) {
	  if(fi > M_PI && fi-FI_STEP < M_PI) {
	    vu_fi-=M_PI;
	    spherical2cartesian(1, M_PI/2, vu_fi, &viewup);
	  }
	  fi-=FI_STEP;
	  if(fi <= 0) {
	    fi = 2*M_PI-FI_STEP;
	    vu_fi+=M_PI;
	    spherical2cartesian(1, M_PI/2, vu_fi, &viewup);
	  }
	}
	XClearWindow(mydisp, mywin);
	draw();
	oldx = tmpx;
	oldy = tmpy;
      }
      break;
    case ButtonPress:
      if(myevent.xbutton.button == 1) 
	button=1;
      if(myevent.xbutton.button == 3) {       /* animazione */
	pseudoteta=teta;
	pseudofi=fi;
	for(;teta<=pseudoteta+2*M_PI; teta+=R_STEP) {
	  if(fi < M_PI && fi+FI_STEP > M_PI) {
	    vu_fi+=M_PI;
	    spherical2cartesian(1, M_PI/2, vu_fi, &viewup);
	  }
	  fi+=FI_STEP;
	  if(fi >= 2*M_PI) {
	    fi=FI_STEP;
	    vu_fi+=M_PI;
	    spherical2cartesian(1, M_PI/2, vu_fi, &viewup);
	  }
	  XClearWindow(mydisp, mywin);
	  draw();
	}
	teta=pseudoteta;
	fi=pseudofi;
      }
      else if(myevent.xbutton.button == 4) {       /* modifica view-up vector */
	vu_fi-=VU_STEP;
	spherical2cartesian(1, M_PI/2, vu_fi, &viewup);
	XClearWindow(mydisp, mywin);
	draw();
      }
      else if(myevent.xbutton.button == 5) {
	vu_fi+=VU_STEP;
	spherical2cartesian(1, M_PI/2, vu_fi, &viewup);
	XClearWindow(mydisp, mywin);
	draw();
      }
    case KeyPress:
      if(myevent.xkey.keycode == 100) {      /* frecce */
	XClearWindow(mydisp,mywin);
	ssx=-sin(teta);
	ssy=cos(teta);
	csx-=sstep*ssx;
	csy-=sstep*ssy;
	draw();
      }
      else if(myevent.xkey.keycode == 102) {
	XClearWindow(mydisp,mywin);
	ssx=-sin(teta);
	ssy=cos(teta);
	csx+=sstep*ssx;
	csy+=sstep*ssy;
	draw();
      }
      else if(myevent.xkey.keycode == 104) {
	XClearWindow(mydisp,mywin);
	ssx=-cos(fi)*cos(teta);
	ssy=-cos(fi)*sin(teta);
	ssz=sin(fi);
	csx+=sstep*ssx;
	csy+=sstep*ssy;
	csz+=sstep*ssz;
	draw();
      }
      else if(myevent.xkey.keycode == 98) {
	XClearWindow(mydisp,mywin);
	ssx=-cos(fi)*cos(teta);
	ssy=-cos(fi)*sin(teta);
	ssz=sin(fi);
	csx-=sstep*ssx;
	csy-=sstep*ssy;
	csz-=sstep*ssz;
	draw();
      }
      else {
	if(XLookupString((XKeyEvent *)&myevent, keyseq, 10, &mykey, 0) > 0) {
	  if(keyseq[0] == 'q') done=1; /* quit */
	  else if(keyseq[0] == 's') { /* + D */
	    D+=dstep;
	    XClearWindow(mydisp, mywin);
	    draw();
	  }
	  else if(keyseq[0] == 'a') { /* - D */
	    if(D-dstep > 0)
	      D-=dstep;
	    XClearWindow(mydisp, mywin);
	    draw();
	  }
	  else if(keyseq[0] == 'd') { /* - D */
	    teta+=TETA_STEP;
	    XClearWindow(mydisp, mywin);
	    draw();
	  }
	  else if(keyseq[0] == 'g') { /* - D */
	    teta-=TETA_STEP;
	    XClearWindow(mydisp, mywin);
	    draw();
	  }
	  else if(keyseq[0] == 'f') { /* - D */
	    if(fi < M_PI && fi+FI_STEP > M_PI) {
	      vu_fi+=M_PI;
	      spherical2cartesian(1, M_PI/2, vu_fi, &viewup);
	    }
	    fi+=FI_STEP;
	    if(fi >= 2*M_PI) {
	      fi=FI_STEP;
	      vu_fi+=M_PI;
	      spherical2cartesian(1, M_PI/2, vu_fi, &viewup);
	    }
	    XClearWindow(mydisp, mywin);
	    draw();
	  }
	  else if(keyseq[0] == 'r') { /* - D */
	    if(fi > M_PI && fi-FI_STEP < M_PI) {
	      vu_fi-=M_PI;
	      spherical2cartesian(1, M_PI/2, vu_fi, &viewup);
	    }
	    fi-=FI_STEP;
	    if(fi <= 0) {
	      fi = 2*M_PI-FI_STEP;
	      vu_fi+=M_PI;
	      spherical2cartesian(1, M_PI/2, vu_fi, &viewup);
	    }
	    XClearWindow(mydisp, mywin);
	    draw();
	  }
	  else if(keyseq[0] == '-') { /* zoom out */
	    Sx-=ZOOM_STEP;
	    Sy-=ZOOM_STEP;
	    if(Sx < 0) Sx=0;
	    if(Sy < 0) Sy=0;
	    XClearWindow(mydisp,mywin);	
	    draw();
	  }
	  else if(keyseq[0] == '+') { /* zoom in */
	    Sx+=ZOOM_STEP;
	    Sy+=ZOOM_STEP;
	    XClearWindow(mydisp,mywin);
	    draw();
	  }
	  else if(keyseq[0] == '1') { /* wireframe */
	    viewtype=WIREFRAME;
	    XSetForeground(mydisp, mygc, myfore);
	    XClearWindow(mydisp,mywin);
	    draw();
	  }
	  else if(keyseq[0] == '2') { /* depth-cuing */
	    viewtype=DEPTH_CUING;
	    XClearWindow(mydisp,mywin);
	    draw();
	  }
	  break; 
	}
      }
    }
  }
  
  XFreeGC(mydisp,mygc);
  XDestroyWindow(mydisp,mywin);
  XCloseDisplay(mydisp);  
  
  return 0;
} 

float sqr(float x)
{
  return x*x;
}
        
float mod(vertex_t *v) 
{
  return sqrt(sqr(v->x)+sqr(v->y)+sqr(v->z));
}

/* divisione di un vettore per uno scalare */
void vecdiv(vertex_t *vin, float scalar, vertex_t *vout) 
{
  vout->x=vin->x/scalar;
  vout->y=vin->y/scalar;
  vout->z=vin->z/scalar;
} 

/* trasformazione da coordinate sferiche a cartesiane */
void spherical2cartesian(float D, float teta, float fi, vertex_t *vout) 
{
  vout->x=D*cos(teta)*sin(fi);
  vout->y=D*sin(teta)*sin(fi);
  vout->z=D*cos(fi);
}

/* trasformazione da coordinate cartesiane a sferiche */
void cartesian2spherical(vertex_t *vin, float *D, float *teta, float *fi) 
{
  float di;

  di = mod(vin);

  *teta = atan2(vin->y, vin->x);
  *fi   = acos(vin->z / di);
  *D    = di;
}

/* arrotondamento */
int round(float g)
{
  if ((g - (int)g) > 0.5) return((int)g + 1);
  return (int)g;
}

/* prodotto scalare */
void dotprod(vertex_t *vin, float matrix[4][4], vertex_t *vout) 
{
  vout->x=vin->x*matrix[0][0] + vin->y*matrix[1][0] + vin->z*matrix[2][0] + matrix[3][0];
  vout->y=vin->x*matrix[0][1] + vin->y*matrix[1][1] + vin->z*matrix[2][1] + matrix[3][1];
  vout->z=vin->x*matrix[0][2] + vin->y*matrix[1][2] + vin->z*matrix[2][2] + matrix[3][2];
}

/* prodotto vettoriale */
void crossprod(vertex_t *v1, vertex_t *v2, vertex_t *vout)
{
  vout->x=v1->y * v2->z - v1->z * v2->y;
  vout->y=v1->z * v2->x - v1->x * v2->z;
  vout->z=v1->x * v2->y - v1->y * v2->x;
}

int det(float m[3][3]) 
{
  int d;

  d=m[0][0]*m[1][1]*m[2][2] + m[0][1]*m[1][2]*m[2][0] + m[0][2]*m[1][0]*m[2][1] + 
    m[2][0]*m[1][1]*m[0][2] + m[2][0]*m[1][2]*m[0][0] + m[2][2]*m[1][0]*m[0][1];

  return d;
}

float zminval(face_t f) 
{
  float t=effe;
  
  if(evert[f.v1].z < t) t=evert[f.v1].z;
  if(evert[f.v2].z < t) t=evert[f.v2].z;
  if(evert[f.v3].z < t) t=evert[f.v3].z;

  return t;
}

/* grubba l'oggetto dalla mesh in formato .m */
void load_object(void)
{
  FILE *f;
  char num[50], vdone=0, fdone=0, edone=0;
  
  if((f=fopen(filename, "r")) == NULL) {
    perror("fopen error");
    exit(-1);
  }
  nvert=0;
  nface=0;
  nedge=0;
  
  while(1) {
    do {
      if(fscanf(f, "%s ", num) == EOF) { vdone=1; break; }
    } while(strcmp(num, "Vertex"));
    if(vdone) break;
    vertex=(vertex_t *)realloc(vertex, (nvert+1)*sizeof(vertex_t));
    if(fscanf(f, "%s  %f %f %f\n", num, &vertex[nvert].x, &vertex[nvert].y, &vertex[nvert].z) == EOF) break;
    vertex[nvert].t=0;
    //printf("Vertex %s  %f %f %f\n", num, vertex[nvert].x, vertex[nvert].y, vertex[nvert].z);
    nvert++;
  }
  if(!vdone) {
    fprintf(stderr, "Cannot load vertexes: syntax error on %s\n", filename);
    exit(-1);
  }
  
  fb=(point_t *)calloc(nvert, sizeof(point_t));
  zbuf=(float *)calloc(nvert, sizeof(float));
  evert=(vertex_t *)calloc(nvert, sizeof(vertex_t));
  
  fseek(f, 0, SEEK_SET); 
  
  while(1) { 
    do { 
      if(fscanf(f, " %s", num) == EOF) { fdone=1; break; } 
    } while(strcmp(num, "Face")); 
    if(fdone) break; 
    face=(face_t *)realloc(face, sizeof(face_t)*(nface+1)); 
    if(fscanf(f, "%s  %d %d %d\n", num, &face[nface].v1, &face[nface].v2, &face[nface].v3) == EOF) break; 
    face[nface].v1--; 
    face[nface].v2--; 
    face[nface].v3--; 
    face[nface].flag=0;
    nface++; 
  } 
  if(!fdone) {
    fprintf(stderr, "Cannot load faces: syntax error on %s\n", filename); 
    exit(-1); 
  } 
  
  fseek(f, 0, SEEK_SET);
  
  while(1) {
    do {
      if(fscanf(f, "%s", num) == EOF) { edone=1; break; }
    } while(strcmp(num, "Edge"));
    if(edone) break;
    edge=(edge_t *)realloc(edge, sizeof(edge_t)*(nedge+1));
    if(fscanf(f, " %d %d %s\n", &edge[nedge].p1, &edge[nedge].p2, num) == EOF) break;
    if(strcmp("{sharp}", num) != 0) { /* alcune mesh descrivono i lati in modo diverso */
      edge[nedge].p1=edge[nedge].p2;
      edge[nedge].p2=atof(num);
    }
    edge[nedge].p1--;
    edge[nedge].p2--;       
    nedge++;
  }
       
  if(!edone) {
    fprintf(stderr, "Cannot load edges: syntax error on %s\n", filename);
    exit(-1);
  }

  fflush(stdout);
}		

/* trasformazione prospettica di un generico punto di vista con traslazione e rotazione */
void world2view_normal(vertex_t *vin, vertex_t *vout)
{
  vertex_t tmp;
  float center[4][4] = {
    {    1,    0,    0,    0 },
    {    0,    1,    0,    0 },
    {    0,    0,    1,    0 },
    { -csx, -csy, -csz,    1 }
  };
  float matrix[4][4] = {
    { -sin(teta), -cos(teta)*cos(fi), -cos(teta)*sin(fi), 0 },
    {  cos(teta), -sin(teta)*cos(fi), -sin(teta)*sin(fi), 0 },
    {          0,            sin(fi),           -cos(fi), 0 },
    {          0,                  0,                  D, 1 }
  };

  dotprod(vin, center, &tmp); 
  dotprod(&tmp, matrix, vout);
}

/* trasformazione prospettica del punto di vista tramite cambiamento di base e view-up vector */
void world2view(vertex_t *vin, float matrix[4][4], vertex_t *vout)
{
  vertex_t tmp;
  float center[4][4] = {
    {    1,    0,    0,    0 },
    {    0,    1,    0,    0 },
    {    0,    0,    1,    0 },
    { -csx, -csy, -csz,    1 }
  };

  dotprod(vin, center, &tmp); 
  dotprod(&tmp, matrix, vout);
}

/* costruzione della matrice per il cambiamento di base */
void make_perspmatrix(float D, float teta, float fi, float matrix[4][4])
{
  vertex_t xe, ye, ze, vp, vd, tmp;

  spherical2cartesian(D, teta, fi, &vd);

  vp.x=-vd.x;
  vp.y=-vd.y;
  vp.z=-vd.z;

  vecdiv(&vd, mod(&vp), &ze);
  
  crossprod(&ze, &viewup, &tmp);
  vecdiv(&tmp, mod(&tmp), &xe);

  crossprod(&ze, &xe, &tmp);
  vecdiv(&tmp, mod(&tmp), &ye);
  
  ye.x=-ye.x;
  ye.y=-ye.y;
  ye.z=-ye.z;

  matrix[0][0]=xe.x;
  matrix[1][0]=xe.y;
  matrix[2][0]=xe.z;
  matrix[3][0]=0.0;
  matrix[0][1]=ye.x;
  matrix[1][1]=ye.y;
  matrix[2][1]=ye.z;
  matrix[3][1]=0.0;
  matrix[0][2]=ze.x;
  matrix[1][2]=ze.y;
  matrix[2][2]=ze.z;
  matrix[3][2]=D;
  matrix[0][3]=0.0;
  matrix[1][3]=0.0;
  matrix[2][3]=0.0;
  matrix[3][3]=1.0;
}

/* funzione che calcola i fattori di scala */
void fattori_scala(void)
{
  int 	k;
  float r;
  float xmin,xmax,ymin,ymax, zmin, zmax;

  if (first) { 
    first=0;
    xmin = vertex[0].x;
    ymin = vertex[0].y;
    zmin = vertex[0].z;
    xmax = xmin;
    ymax = ymin;
    zmax = zmin;
    for (k=1;k<nvert;k++) {
      if (vertex[k].x > xmax) xmax = vertex[k].x;
      else if (vertex[k].x < xmin) xmin = vertex[k].x;
      if (vertex[k].y > ymax) ymax = vertex[k].y;
      else if (vertex[k].y < ymin) ymin = vertex[k].y;
      if (vertex[k].z > zmax) zmax = vertex[k].z;
      else if (vertex[k].z < zmin) zmin = vertex[k].z;
    }
    /* centro oggetto iniziale */
    csx = (xmin + xmax)/2;
    csy = (ymin + ymax)/2;
    csz = (zmin + zmax)/2;
    dstep = D/100;
    sstep = (xmax-xmin)/50;
    /* raggio sfera contenete oggetto */
    r = sqrt(sqr(xmax-csx)+sqr(ymax-csy)+sqr(zmax-csz));
    /* calcolo semilato window iniziale; il piano di proiezione viene assunto
       ad una distanza D dall'osservatore */
    semw = r*di/sqrt(sqr(di)-sqr(r));
    /* in base al semilato window cosi' calcolato si determina l'apertura
       angolare iniziale */
    alpha=atan(semw/di);
    /* printf("\nApertura angolare: %f \n",alpha);*/
    /* fattori di scala per trasf. Window-Viewport */
    xwmin = -semw ; 
    xwmax = semw;
    ywmin = -semw; 
    ywmax = semw;
    Sx = ((float)(xvmax) - (float)(xvmin))/(xwmax - xwmin);
    Sy = ((float)(yvmax) - (float)(yvmin))/(ywmax - ywmin);
  } 
  else { 
    /* si ricalcola il semilato della window in base a D e ad alpha */
    semw=di*tan(alpha);
  }
  xwmin = -semw ; 
  xwmax = semw;
  ywmin = -semw; 
  ywmax = semw;
}

/* clipping basato su Cohen-Southerland tridimensionalizzato (incompleto) */
void clip(float matrix[4][4]) 
{
  float homogeneous[4][4] = {
    { di/semw,       0, 0, 0 },
    {       0, di/semw, 0, 0 },
    {       0,       0, 1, 0 },
    {       0,       0, 0, 1 }
  };
  vertex_t vf, *v1, *v2, *v3;
  int k, vpoint1, vpoint2, vpoint3, pos1, pos2, pos3;

  for(k=0; k<nface; k++) {
    pos1=0;
    pos2=0;
    pos3=0;
    vpoint1=face[k].v1;
    vpoint2=face[k].v2;
    vpoint3=face[k].v3;
    v1=&vertex[vpoint1];
    v2=&vertex[vpoint2];
    v3=&vertex[vpoint3];
    world2view(v1, matrix, &vf);
    dotprod(&vf, homogeneous, &vf);
    if(vf.z < di)       pos1 |= 0x01;
    if(vf.z > effe) 	pos1 |= 0x02;
    if(vf.x > vf.z)	pos1 |= 0x04;
    if(vf.x < -vf.z) 	pos1 |= 0x08;
    if(vf.y > vf.z) 	pos1 |= 0x10;
    if(vf.y < -vf.z)    pos1 |= 0x20;

    world2view(v2, matrix, &vf);
    dotprod(&vf, homogeneous, &vf);
    if(vf.z < di)       pos2 |= 0x01;
    if(vf.z > effe) 	pos2 |= 0x02;
    if(vf.x > vf.z)	pos2 |= 0x04;
    if(vf.x < -vf.z) 	pos2 |= 0x08;
    if(vf.y > vf.z) 	pos2 |= 0x10;
    if(vf.y < -vf.z)    pos2 |= 0x20;

    world2view(v3, matrix, &vf);
    dotprod(&vf, homogeneous, &vf);
    if(vf.z < di)       pos3 |= 0x01;
    if(vf.z > effe) 	pos3 |= 0x02;
    if(vf.x > vf.z)	pos3 |= 0x04;
    if(vf.x < -vf.z) 	pos3 |= 0x08;
    if(vf.y > vf.z) 	pos3 |= 0x10;
    if(vf.y < -vf.z)    pos3 |= 0x20;

    if((pos1 | pos2) == 0) {
      v1->t=INSIDE;
      v2->t=INSIDE;
    }
    else if(pos1 & pos2) {
      v1->t=CLIPPED;
      v2->t=CLIPPED;
    }
    else {
      if(pos1 != 0 && pos2 != 0) {
	// entrambi i punti sono fuori ma da zone diverse (per ora CLIPPO)
	v1->t=CLIPPED;
	v2->t=CLIPPED;
      }
      else {
	// uno dei due punti e' esterno, da calcolare l'intersezione con il piano (incompleto)
	v1->t=INSIDE;
	v2->t=INSIDE;
      }      
    }
    
    if((pos1 | pos3) == 0) {
      v1->t=INSIDE;
      v3->t=INSIDE;
    }
    else if(pos1 & pos3) {
      v1->t=CLIPPED;
      v3->t=CLIPPED;
    }
    else {
      if(pos1 != 0 && pos3 != 0) {
	v1->t=CLIPPED;
	v3->t=CLIPPED;
      }
      else {
	v1->t=INSIDE;
	v3->t=INSIDE;
      }
    }      

    if((pos2 | pos3) == 0) {
      v2->t=INSIDE;
      v3->t=INSIDE;
    }
    else if(pos2 & pos2) {
      v2->t=CLIPPED;
      v3->t=CLIPPED;
    }
    else {
      if(pos2 != 0 && pos3 != 0) {
	v2->t=CLIPPED;
	v3->t=CLIPPED;
      }
      else {
	v2->t=INSIDE;
	v3->t=INSIDE;
      }
    }      
  }

  for(k=0; k<nedge; k++) {
    pos1=0;
    pos2=0;
    vpoint1=edge[k].p1;
    vpoint2=edge[k].p2;
    v1=&vertex[vpoint1];
    v2=&vertex[vpoint2];
    world2view(v1, matrix, &vf);
    dotprod(&vf, homogeneous, &vf);
    if(vf.z < di)       pos1 |= 0x01;
    if(vf.z > effe) 	pos1 |= 0x02;
    if(vf.x > vf.z)	pos1 |= 0x04;
    if(vf.x < -vf.z) 	pos1 |= 0x08;
    if(vf.y > vf.z) 	pos1 |= 0x10;
    if(vf.y < -vf.z)    pos1 |= 0x20;

    world2view(v2, matrix, &vf);
    dotprod(&vf, homogeneous, &vf);
    if(vf.z < di)       pos2 |= 0x01;
    if(vf.z > effe) 	pos2 |= 0x02;
    if(vf.x > vf.z)	pos2 |= 0x04;
    if(vf.x < -vf.z) 	pos2 |= 0x08;
    if(vf.y > vf.z) 	pos2 |= 0x10;
    if(vf.y < -vf.z)    pos2 |= 0x20;

    if((pos1 | pos2) == 0) {
      v1->t=INSIDE;
      v2->t=INSIDE;
    }
    else if(pos1 & pos2) {
      v1->t=CLIPPED;
      v2->t=CLIPPED;
    }
    else {
      if(pos1 != 0 && pos2 != 0) {
	v1->t=CLIPPED;
	v2->t=CLIPPED;
      }
      else {
	v1->t=INSIDE;
	v2->t=INSIDE;
      }
    }
  }
}

void createcolors(void) 
{
  XColor col;

  for(i=0; i<256; i++) {
    /* tolgo i colori troppo scuri uniformandoli tutti al 32esimo grigio */
    if(i < 32) col.red = col.green = col.blue = 32*256;
    else       col.red = col.green = col.blue = i*256;
    XAllocColor(mydisp, cmap, &col);
    scale[i]=col.pixel;
  }
}

/* ... */
void draw(void)
{
  switch(viewtype) {
  case WIREFRAME:
    draw_wireframe();
    break;
  case DEPTH_CUING:
    draw_depthcuing();
    break;
  }
}

void draw_wireframe(void) 
{    
  int t, u, k, clipped;
  vertex_t ve;
  char values[100];
  float vu_matrix[4][4];
  
  if (first) 
    load_object();   /* determinazione mesh oggetto */

  fattori_scala(); /* calcolo dei fattori di scala */

  make_perspmatrix(D, teta, fi, vu_matrix);

  /* clipping (incompleto) */
  clip(vu_matrix);

  for (k=0; k<nvert; k++) {
    world2view(&vertex[k], vu_matrix, &ve); 
    /* proiezione e trasformazione in coordinate schermo */
    fb[k].x = round(Sx * ((di * ve.x)/ve.z - (xwmax+xwmin)/2) + (xvmax+xvmin)/2);
    fb[k].y = round(Sy * ((ywmax+ywmin)/2 - (di * ve.y)/ve.z) + (yvmax+yvmin)/2);
  }

  clipped=0;

  /* disegno mesh */
  for (k=0; k<nface; k++) {
    t=face[k].v1;
    u=face[k].v2;

    if(vertex[t].t == INSIDE && vertex[u].t == INSIDE) 
      XDrawLine(mydisp, mywin, mygc, fb[t].x, fb[t].y, fb[u].x, fb[u].y);
    else 
      clipped++;

    t=face[k].v1;
    u=face[k].v3;
    if(vertex[t].t == INSIDE && vertex[u].t == INSIDE) 
      XDrawLine(mydisp, mywin, mygc, fb[t].x, fb[t].y, fb[u].x, fb[u].y);
    else
      clipped++;

    t=face[k].v2;
    u=face[k].v3;
    if(vertex[t].t == INSIDE && vertex[u].t == INSIDE) 
      XDrawLine(mydisp, mywin, mygc, fb[t].x, fb[t].y, fb[u].x, fb[u].y);
    else
      clipped++;
  }
  
  for (k=0; k<nedge; k++) {
    t=edge[k].p1;
    u=edge[k].p2;
    if(vertex[t].t == INSIDE && vertex[u].t == INSIDE) 
      XDrawLine(mydisp, mywin, mygc, fb[t].x, fb[t].y, fb[u].x, fb[u].y);
    else
      clipped++;
  }

  snprintf(values, 100, "[ D = %3.2f ] [ teta = %d ] [ fi = %d ]", D, round(teta*180/M_PI)%360, round(fi*180/M_PI)%360);
  XDrawImageString(mydisp, mywin, mygc, 2, 15, values, strlen(values));
  snprintf(values, 100, "[ di = %3.0f ] [ f = %3.0f ] [ vu_fi = %d ]", di, effe, round(vu_fi*180/M_PI)%360);
  XDrawImageString(mydisp, mywin, mygc, 2, 30, values, strlen(values));
  snprintf(values, 100, "[ V = %d ] [ F = %d ] [ E = %d ]", nvert, nface, nedge);
  XDrawImageString(mydisp, mywin, mygc, 2, 45, values, strlen(values));
  if(clipped > 0) {
    snprintf(values, 100, "[ Clipped = %d ]", clipped);
    XDrawImageString(mydisp, mywin, mygc, 2, 60, values, strlen(values));
  }    
}				

void draw_depthcuing(void) 
{    
  int t, u, k, clipped;
  vertex_t ve;
  char values[100];
  float vu_matrix[4][4], zcolor, zmax=0, zmin=effe;
  
  if (first) 
    load_object();   /* determinazione mesh oggetto */

  fattori_scala(); /* calcolo dei fattori di scala */

  make_perspmatrix(D, teta, fi, vu_matrix);

  /* clipping (incompleto) */
  clip(vu_matrix);

  for (k=0; k<nvert; k++) {
    world2view(&vertex[k], vu_matrix, &ve); 
    //world2view_normal(&vertex[k], &ve);
    zbuf[k] = ve.z;
    zmax = ve.z > zmax ? ve.z : zmax;
    zmin = ve.z < zmin ? ve.z : zmin;
      
    /* proiezione e trasformazione in coordinate schermo */
    fb[k].x = round(Sx * ((di * ve.x)/ve.z - (xwmax+xwmin)/2) + (xvmax+xvmin)/2);
    fb[k].y = round(Sy * ((ywmax+ywmin)/2 - (di * ve.y)/ve.z) + (yvmax+yvmin)/2);
  }

  clipped=0;

  /* disegno mesh */
  for (k=0; k<nface; k++) {
    t=face[k].v1;
    u=face[k].v2;

    if(vertex[t].t == INSIDE && vertex[u].t == INSIDE) {
      zcolor=(((zbuf[t] + zbuf[u])/2)-zmin)/(zmax - zmin);
      XSetForeground(mydisp, mygc, scale[(int)(255*(1-zcolor))]);
      XDrawLine(mydisp, mywin, mygc, fb[t].x, fb[t].y, fb[u].x, fb[u].y);
    }
    else 
      clipped++;

    t=face[k].v1;
    u=face[k].v3;
    if(vertex[t].t == INSIDE && vertex[u].t == INSIDE) {
      zcolor=(((zbuf[t] + zbuf[u])/2)-zmin)/(zmax - zmin);
      XSetForeground(mydisp, mygc, scale[(int)(255*(1-zcolor))]);
      XDrawLine(mydisp, mywin, mygc, fb[t].x, fb[t].y, fb[u].x, fb[u].y);
    }
    else
      clipped++;

    t=face[k].v2;
    u=face[k].v3;
    if(vertex[t].t == INSIDE && vertex[u].t == INSIDE) {
      zcolor=(((zbuf[t] + zbuf[u])/2)-zmin)/(zmax - zmin);
      XSetForeground(mydisp, mygc, scale[(int)(255*(1-zcolor))]);
      XDrawLine(mydisp, mywin, mygc, fb[t].x, fb[t].y, fb[u].x, fb[u].y);
    }
    else
      clipped++;
  }
  
  for (k=0; k<nedge; k++) {
    t=edge[k].p1;
    u=edge[k].p2;
    if(vertex[t].t == INSIDE && vertex[u].t == INSIDE) {
      zcolor=(((zbuf[t] + zbuf[u])/2)-zmin)/(zmax - zmin);
      XSetForeground(mydisp, mygc, scale[(int)(255*(1-zcolor))]);
      XDrawLine(mydisp, mywin, mygc, fb[t].x, fb[t].y, fb[u].x, fb[u].y);
    }
    else
      clipped++;
  }

  XSetForeground(mydisp, mygc, myfore);
  snprintf(values, 100, "[ D = %3.2f ] [ teta = %d ] [ fi = %d ]", D, round(teta*180/M_PI)%360, round(fi*180/M_PI)%360);
  XDrawImageString(mydisp, mywin, mygc, 2, 15, values, strlen(values));
  snprintf(values, 100, "[ di = %3.0f ] [ f = %3.0f ] [ vu_fi = %d ]", di, effe, round(vu_fi*180/M_PI)%360);
  XDrawImageString(mydisp, mywin, mygc, 2, 30, values, strlen(values));
  snprintf(values, 100, "[ V = %d ] [ F = %d ] [ E = %d ]", nvert, nface, nedge);
  XDrawImageString(mydisp, mywin, mygc, 2, 45, values, strlen(values));
  if(clipped > 0) {
    snprintf(values, 100, "[ Clipped = %d ]", clipped);
    XDrawImageString(mydisp, mywin, mygc, 2, 60, values, strlen(values));
  }
}			

